/*
 * ATX_pwradapt_basic.cpp
 *
 * Created: 21.8.2013 20:21:01
 * Author: Martin Lukasek (www.8bity.cz)
 */ 


#include <avr/io.h>
#include <avr/interrupt.h> 
#define F_CPU 8000000UL
#include <util/delay.h>

#define TICK_PIN PB0
#define PWROK_PIN PB1
#define MEASURE_PIN PB2
#define PWRON_PIN PB3
#define KEY_PIN PB4
#define LONG_PRESS 20 // 20

#define output_low(port,pin) port &= ~(1<<pin)
#define output_high(port,pin) port |= (1<<pin)
#define set_input(portdir,pin) portdir &= ~(1<<pin)
#define set_output(portdir,pin) portdir |= (1<<pin)
#define toggle(reg,bit)     reg ^= (1<<(bit)) 

int swOn = 0;
int generate_tick50Hz = 0;
int fail_indicator = 0;
int generate_reset = 1;
int check_pwrOK = 0;
int check_5V = 1;
int pwr_err = 1;
unsigned long milliseconds = 0;

void initADC()
{
  // disable digital input on ADC1/PB2 pins
  output_high(DIDR0,ADC1D); 

  /* this function initialises the ADC 
 	 ADC Prescaler needs to be set so that the ADC input frequency is between 50 - 200kHz.
	
	 Clock   Available prescaler values
    ---------------------------------------
 	 1 MHz   8 (125kHz), 16 (62.5kHz) 
	 4 MHz   32 (125kHz), 64 (62.5kHz)
	 8 MHz   64 (125kHz), 128 (62.5kHz)
	16 MHz   128 (125kHz)
  */

  ADMUX =
            (0 << ADLAR) |     // left shift result
            (1 << REFS2) |     // Sets ref. voltage to VCC, bit 2
            (1 << REFS1) |     // Sets ref. voltage to VCC, bit 1
            (0 << REFS0) |     // Sets ref. voltage to VCC, bit 0
            (0 << REFS0) |     // Sets ref. voltage to VCC, bit 0
            (0 << MUX3)  |     // use ADC1 for input (PB2), MUX bit 3
            (0 << MUX2)  |     // use ADC1 for input (PB2), MUX bit 2
            (0 << MUX1)  |     // use ADC1 for input (PB2), MUX bit 1
            (1 << MUX0);       // use ADC1 for input (PB2), MUX bit 0

  ADCSRA = 
            (1 << ADEN)  |     // Enable ADC 
            (1 << ADPS2) |     // set prescaler to 64, bit 2 
            (1 << ADPS1) |     // set prescaler to 64, bit 1 
            (0 << ADPS0);      // set prescaler to 64, bit 0  
}

void initTimer0()
{
	TCCR0B |= ((0 << CS02) | (1 << CS01) | (1 << CS00)); // Timer 0 prescaling - divides by 64 */
	TCCR0A |= (1 << WGM01); // Put timer 0 in CTC mode
	OCR0A = 125; // Count 125 cycles for interrupt
	TIMSK |= (1 << OCIE0A); // enable timer compare interrupt
}

void setSwitch(void)
{
  if(swOn) {
    output_low(PORTB, PWRON_PIN); // Set 0 = POWER ON
  } else {
	output_high(PORTB, PWRON_PIN); // Set 1 = POWER OFF
	output_low(PORTB, TICK_PIN); // Use xor to toggle the K7

  }
}

int keyPress(void) 
{ 
  int presses=0; 
  for(int i=0; i<8; i++) { 
    if( !bit_is_set(PINB, KEY_PIN) ) { // read pin == 1 
      presses++; 
    } 
  _delay_ms(5); 
  } 
  return presses>7; 
}  

ISR(TIMER0_COMPA_vect) // Interrupt Service Routine - 1 ms
{
	milliseconds++;
	switch( pwr_err ) {
		case 0:
			if( swOn && generate_tick50Hz && ((milliseconds % 10) == 0) ) {
				PORTB ^= (1 << TICK_PIN); // Use xor to toggle the K7
			}
			break;
		case 1:
			if( fail_indicator && ((milliseconds % 250) == 0) ) {
				PORTB ^= (1 << TICK_PIN); // Use xor to toggle the K7
			}
			break;
		case 2:
			if( fail_indicator && ((milliseconds % 1000) == 0) ) {
				PORTB ^= (1 << TICK_PIN); // Use xor to toggle the K7
			}
			break;
	}
}

int main(void)
{
	int measured;

	cli(); // Disable global interrupts

    DDRB |= (1 << TICK_PIN); // Set output direction on TICK_PIN
    if( fail_indicator || generate_reset ) {
	    output_low(PORTB, TICK_PIN) ; // keep reset or no-fail while initializing
    }

    DDRB &= ~(1 << KEY_PIN); // Set input direction on KEY_PIN
    PORTB |= (1 << KEY_PIN); // Enable pull up on KEY_PIN

    DDRB &= ~(1 << PWROK_PIN); // Set input direction on PWROK_PIN
    PORTB &= ~(1 << PWROK_PIN); // Disable pull up on PWROK_PIN

    DDRB |= (1 << PWRON_PIN); // Set output direction on LED_PIN
    setSwitch(); // POWER OFF

    initADC();
	// measure once, as first measure does not have to be exact
   	ADCSRA |= (1 << ADSC);         // start ADC measurement
	while (ADCSRA & (1 << ADSC) ); // wait till conversion complete
	
    initTimer0();
	
    sei(); // Enable global interrupts
	
    while(1)
    {
        if ( keyPress() ) {
	        if(swOn) {
		        int cnt=0;
		        while( keyPress() & (cnt<LONG_PRESS) ) { // Wait if key is pressed for longer time
			        _delay_ms(60);
			        cnt++;
		        }
		        if(cnt==LONG_PRESS)	{ // Swith OFF of after long key press
			        swOn = 0;
			        setSwitch();
			        while(keyPress()) {} // Wait for key release
			        _delay_ms(500);
		        } else {
					if( generate_reset && (cnt>0) ) { // short press
					    output_low(PORTB, TICK_PIN) ; // make reset
				        _delay_ms(500);
					    output_high(PORTB, TICK_PIN) ; // make reset
					}
				}
		    } else {
		        swOn = 1; // Swith ON of after key press
		        setSwitch();
				milliseconds = 0; // reset power on timer
				pwr_err = 0; // reset error
		        while(keyPress()) {}  // Wait for key release
		        _delay_ms(500);
	        }
        }

        if( (swOn>0) && (milliseconds>2000) ) { // check power only when ON and more than 2s after ON
			// measure and check -5V
			ADCSRA |= (1 << ADSC);         // start ADC measurement
			while (ADCSRA & (1 << ADSC) ); // wait till conversion complete
			measured = ADCW;
			if( check_5V && ((measured>696) || (measured<468)) ) { // combination of +5V and -5V is worse than 4,5...5,5V
			    swOn = 0; // POWER OFF
			    setSwitch();
				pwr_err = 1; // -5V or +5V out of bounds
			    _delay_ms(1000); // wait for power OFF
			} else {
				if( check_pwrOK && (!bit_is_set(PINB, PWROK_PIN)) ) { // POWER OK signal failed
					swOn = 0; // POWER OFF
					setSwitch();
					pwr_err = 2; // POWER OK signal failed
					_delay_ms(1000); // wait for power OFF
				} else {
				    output_high(PORTB, TICK_PIN) ; // everything is ok - release (or keep released) the RESET
				}
			}
		}
    }
}